/* primes.c */

#include <sys/types.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>

/* numeral_to_int(s, i) tries to convert s to an integer, returning 0
   if it fails, and storing the integer in *i and returning 1, if it
   suceeds.  The numeral must be in base 10, may optionally be
   preceded by "+" to indicate a positive number, or "-" to indicate a
   negative number, and may be preceded and/or followed by
   whitespace */

int numeral_to_int(char *s, int *i) {
  if (*s == '\0') {
    return 0;
  }
  else {
    char *t;
    long n = strtol(s, &t, 10);
    if (*t == '\0') {
      if (n >= INT_MIN && n <= INT_MAX) {
        *i = n; return 1;
      }
      else {
        return 0;
      }
    }
    else {
      return 0;
    }
  }
}

/* if n >= 0 and primes points to n integers, then gen_primes(n,
   primes) generates the first n primes, storing them in primes[0],
   ..., primes[n - 1] */

void gen_primes(int n, int *primes) {
  int i, j;
  int next = 2;  /* next candidate */
  for (i = 0; i < n; i++) {
    int found = 0;
    while (!found) {
      for (j = 0; j < i; j++) {
        int p = primes[j];
        if (next % p == 0) { break; }
        else if (p * p > next) { j = i; break; }
      }
      if (j == i) { found = 1; }
      else { next++; }
    }
    primes[i] = next++;
  }
}

/* if n >= 0, and ints points to n initialized integers, then
   print_ints(n, ints) prints int[0], ..., int[n - 1],
   one numeral per line */

void print_ints(int n, int *ints) {
  int i;
  int *p = ints;
  for (i = 0; i < n; i++) {
    printf("%d\n", *p++);
  }
}

int error(char *prog, char *s) {
  printf("%s: %s\n", prog, s);
  return 1;
}  

int usage(char *prog) {
  printf("%s: usage: %s num\n", prog, prog);
  return 1;
}

int main(int argc, char **argv) {
  int n;
  int *primes;
  if (argc != 2) {
    return usage(argv[0]);
  }
  else if (numeral_to_int(argv[1], &n)) {
    if (n < 0) {
      return error(argv[0], "argument must be nonnegative integer");
    }
    else if ((primes = calloc(n, sizeof(int))) == NULL) {
      return error(argv[0], "storage allocation failed");
    }
    else {
      gen_primes(n, primes);
      print_ints(n, primes);
      return 0;
    }
  }
  else {
    return error(argv[0], "argument must be nonnegative integer");
  }
}
